// qsynthAbout.h
//
/****************************************************************************
   Copyright (C) 2003-2015, rncbc aka Rui Nuno Capela. All rights reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*****************************************************************************/

#ifndef __qsynthAbout_h
#define __qsynthAbout_h

#include "config.h"

#define QSYNTH_TITLE        PACKAGE_NAME
#define QSYNTH_VERSION      PACKAGE_VERSION

#define QSYNTH_SUBTITLE     "A fluidsynth Qt GUI Interface"
#define QSYNTH_WEBSITE      "http://qsynth.sourceforge.net"
#define QSYNTH_COPYRIGHT    "Copyright (C) 2003-2015, rncbc aka Rui Nuno Capela. All rights reserved."

#define QSYNTH_DOMAIN       "rncbc.org"

#endif  // __qsynthAbout_h

// end of qsynthAbout.h
